#pragma once
#include <vector>
#include <vcg/point2.h>
#include <vcg/plane3.h>
#include <vcg/matrix44.h>
#include <QString.h>
#include <QTextStream.h>
#include <qdom.h> 

namespace vcg {

class MMSphere 
{
	public:
		MMSphere():p(0,0,0),v(0,0,0),rad(0.25f){dl=0;};
	float rad;
	Point3f p;
	Point3f v;
	Point3f a;
	GLuint dl;
	virtual ~MMSphere(){};
	virtual void drawGL();
};

class MMFace
{
public:
	MMFace(){};
	MMFace(Point3f * v0, Point3f * v1, Point3f * v2);
	enum {// Face Orientation Flags, used efficiently compute point face distance  
		NORMX		= 0x00000200,
		NORMY		= 0x00000400,
		NORMZ		= 0x00000800,
	};

	Point3f *v[3];
	Point3f n;
	Point3f edge[3];
	Plane3f plane;
	int flags;
   
	void ComputeE();

	// Point face distance
	bool Dist( const Point3f & q, float & dist, Point3f & p );
};

class MMCell 
{
public:
	bool selected;
	Point3f V[4];
	Point3f B[4];
	Point3f n;
	float tmpH[4];
	void ComputeNormal();
	void drawGL();
	float avgH() {return (V[0][2]+V[1][2]+V[2][2]+V[3][2])/4.0f;}
	std::vector<MMFace> face;
	void generateFace();
	void updateSpherePos(MMSphere *sp);
	bool collideSphere(MMSphere *s,Point3f &impact);
 	virtual void XMLWrite(QTextStream &xstrm);
	virtual bool XMLRead(QTextStream &xstrm);
};


class MMGrid 
{
public:
	MMGrid() :ti(0){resize(10,10);}
	MMGrid(const int _sx, const int _sy) {resize(_sx,_sy);}
  GLuint ti;
	void setTexture(QString t) {texFileName=t;};
	void initGL();
	void resize(const int sx, const int sy)  ;
	inline MMCell &G(const int i, const int j){return g[j*sx+i];}
	void drawGL();
	bool pickGL(Point2i in, Point2i &pos);
	bool pickGL(Point3d in, Point2i &pos);
  void clearSelection();
	void selectRect(Point2i s, Point2i e);
 void raiseSelected(float f);
 void smoothSelected();
 void flattenSelected();
 void generateCellVertAndFace();
 void updateSpherePos(float dt);
 void setBallOnSelectionCenter();
 void generateFace();


protected:
	std::vector<MMCell> g;
	int sx,sy;
	QString texFileName;
    
	// for the picking
	double mm[16],mp[16];
	int vp[4]; // the viewport
	MMSphere sp;
};
} //end namespace
